/*--------------------------------------------------------------------*/
/*--- Management of symbols and debugging information.             ---*/
/*---                                                 vg_symtab2.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of Valgrind, an extensible x86 protected-mode
   emulator for monitoring program execution on x86-Unixes.

   Copyright (C) 2000-2004 Julian Seward 
      jseward@acm.org

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#include "vg_include.h"
#include "vg_symtabd.h"

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <elf.h>

static Bool symtabd_initialised = False;
static Int symtabd_send = -1;
static Int symtabd_recv = -1;

/* A structure which contains information pertaining to one mapped
   text segment. (typedef in vg_skin.h) */
struct _SegInfo {
   struct _SegInfo* next;	/* list of SegInfos */

   Segment	*seg;		/* first segment we're mapped out of */
   Int		ref;

   /* Description of the mapped segment. */
   Addr   start;
   UInt   size;
   Char*  filename; /* in mallocville */
   UInt   foffset;
   Char*  soname;

   UInt	  handle;
};

static SegInfo* segInfo = NULL;

static void
checkRead(Int fd, void* buf, size_t len, const char* s)
{
   if (VG_(read)(fd, buf, len) != len) {
      char msg[256];
      VG_(sprintf)(msg, "Error reading '%s'", s);
      VG_(close)(symtabd_send);
      VG_(close)(symtabd_recv);
      symtabd_send = -1;
      symtabd_recv = -1;
      VG_(core_panic)(msg);
   }
}

static void
checkWrite(Int fd, const void* buf, size_t len, const char* s)
{
   if (VG_(write)(fd, buf, len) != len) {
      char msg[256];
      VG_(sprintf)(msg, "Error writing '%s'", msg);
      VG_(close)(symtabd_send);
      VG_(close)(symtabd_recv);
      symtabd_send = -1;
      symtabd_recv = -1;
      VG_(core_panic)(msg);
   }
}

static void freeSegInfo ( SegInfo* si )
{
   vg_assert(si != NULL);
   if (si->filename) VG_(arena_free)(VG_AR_SYMTAB, si->filename);

   if (symtabd_send != -1) {
      Char cmd = VG_SYMTABD_FREE_SYMBOLS;
      checkWrite(symtabd_send, &cmd, 1, "command");
      checkWrite(symtabd_send, &si->handle, sizeof(si->handle),
		 "symtab handle");
   }

   VG_(arena_free)(VG_AR_SYMTAB, si);
}

Bool
VG_(is_object_file)( const void *buf )
{
   {
      Elf32_Ehdr *ehdr = (Elf32_Ehdr *)buf;
      Int ok = 1;

      ok &= (ehdr->e_ident[EI_MAG0] == 0x7F
             && ehdr->e_ident[EI_MAG1] == 'E'
             && ehdr->e_ident[EI_MAG2] == 'L'
             && ehdr->e_ident[EI_MAG3] == 'F');
      ok &= (ehdr->e_ident[EI_CLASS] == ELFCLASS32
             && ehdr->e_ident[EI_DATA] == ELFDATA2LSB
             && ehdr->e_ident[EI_VERSION] == EV_CURRENT);
      ok &= (ehdr->e_type == ET_EXEC || ehdr->e_type == ET_DYN);
      ok &= (ehdr->e_machine == EM_386);
      ok &= (ehdr->e_version == EV_CURRENT);
      ok &= (ehdr->e_shstrndx != SHN_UNDEF);
      ok &= (ehdr->e_shoff != 0 && ehdr->e_shnum != 0);
      ok &= (ehdr->e_phoff != 0 && ehdr->e_phnum != 0);

      if (ok)
	 return True;
   }

   /* other file formats here? */

   return False;
}

/* Print into buf info on code address, function name and filename */
Char* VG_(describe_eip)(Addr eip, Char* buf, Int n_buf)
{
#define APPEND(str)                                         \
   { UChar* sss;                                            \
     for (sss = str; n < n_buf-1 && *sss != 0; n++,sss++)   \
        buf[n] = *sss;                                      \
     buf[n] = '\0';                                         \
   }
   UInt  lineno; 
   UChar ibuf[20];
   UInt  n = 0;
   UChar buf_fn[M_VG_ERRTXT];
   UChar buf_obj[M_VG_ERRTXT];
   UChar buf_srcloc[M_VG_ERRTXT];
   Bool  know_fnname  = VG_(get_fnname) (eip, buf_fn,  M_VG_ERRTXT);
   Bool  know_objname = VG_(get_objname)(eip, buf_obj, M_VG_ERRTXT);
   Bool  know_srcloc  = VG_(get_filename_linenum)(eip, buf_srcloc, M_VG_ERRTXT, 
                                                  &lineno);
   VG_(sprintf)(ibuf,"0x%x: ", eip);
   APPEND(ibuf);
   if (know_fnname) { 
      APPEND(buf_fn);
      if (!know_srcloc && know_objname) {
         APPEND(" (in ");
         APPEND(buf_obj);
         APPEND(")");
      }
   } else if (know_objname && !know_srcloc) {
      APPEND("(within ");
      APPEND(buf_obj);
      APPEND(")");
   } else {
      APPEND("???");
   }
   if (know_srcloc) {
      APPEND(" (");
      APPEND(buf_srcloc);
      APPEND(":");
      VG_(sprintf)(ibuf,"%d",lineno);
      APPEND(ibuf);
      APPEND(")");
   }
   return buf;

#undef APPEND
}

void
VG_(mini_stack_dump)( Addr eips[], UInt n_eips )
{
   UInt  i;
   UChar buf[M_VG_ERRTXT];
   Bool  main_done = False;

   Int stop_at = n_eips;

   vg_assert(stop_at > 0);

   /* This loop is the basis for the one in VG_(gen_suppressions)();  if you
      change this, change it too! */
   i = 0;
   do {
      Addr eip = eips[i];
      if (i  > 0) eip--;            /* point to calling line */
      VG_(describe_eip)(eip, buf, M_VG_ERRTXT);

      if ( ! VG_(clo_show_below_main)) {
         // Stop after "main";  if main() is recursive, stop after last main().
         if (VG_(strstr)(buf, " main ("))
            main_done = True;
         else if (main_done)
            break;
      }
      VG_(message)(Vg_UserMsg, "   %s %s", ( i == 0 ? "at" : "by" ), buf);
      i++;

   } while (i < (UInt)stop_at && eips[i] != 0);
}

SegInfo *
VG_(read_seg_symbols)( Segment *seg )
{
   SegInfo* si;

   if (!symtabd_initialised) {
      Int send_fds[2];
      Int recv_fds[2];
      char send_buf[10];
      char recv_buf[10];

      //VG_(printf)("starting vg_symtabd\n");
      VG_(pipe)(send_fds);
      VG_(pipe)(recv_fds);

      symtabd_send = VG_(safe_fd)(send_fds[0]);
      symtabd_recv = VG_(safe_fd)(recv_fds[1]);

      if (fork() == 0) {
	 char* valgrind_lib = VG_LIBDIR;
	 char* p = getenv(VALGRINDLIB);
	 char path[PATH_MAX];
	 char* args[4];

	 if (p) valgrind_lib = p;

	 VG_(sprintf)(path, "%s/vg_symtabd", valgrind_lib);
	 //VG_(printf)("Executing %s\n", path);
	 VG_(sprintf)(send_buf, "%d", send_fds[1]);
	 VG_(sprintf)(recv_buf, "%d", recv_fds[0]);
	 args[0] = "vg_symtabd";
	 args[1] = send_buf;
	 args[2] = recv_buf;
	 args[3] = 0;
	 execve(path, args, 0);
	 VG_(printf)("Can't execute %s\n", path);
	 VG_(exit)(1);
      } else {
	 Char cmd = VG_SYMTABD_INIT;

	 VG_(close)(recv_fds[0]);
	 VG_(close)(send_fds[1]);

	 checkWrite(symtabd_send, &cmd, 1, "command");
	 checkWrite(symtabd_send, &VG_(clo_verbosity),	sizeof(Int),
		    "clo_verbosity");
	 checkWrite(symtabd_send, &VG_(clo_profile),	sizeof(Bool),
		    "clo_profile");
	 checkWrite(symtabd_send, &VG_(clo_show_below_main), sizeof(Bool),
		    "clo_show_below_main");
	 checkWrite(symtabd_send, &VG_(clo_trace_symtab), sizeof(Bool),
		    "clo_trace_symtab");
	 checkWrite(symtabd_send, &VG_(clo_logfile_fd),	sizeof(Int),
		    "clo_logfile_fd");
	 checkWrite(symtabd_send, &VG_(needs),		sizeof(VgNeeds),
		    "needs");
      }

      symtabd_initialised = True;
   }

   //VG_(printf)("read_seg_symbols(%s, %p, %d)\n", seg->filename, seg->addr, seg->len);

   vg_assert(seg->symtab == NULL);

   VGP_PUSHCC(VgpReadSyms);

   /* Get the record initialised right. */
   si = VG_(arena_malloc)(VG_AR_SYMTAB, sizeof(SegInfo));

   VG_(memset)(si, 0, sizeof(*si));
   si->start    = seg->addr;
   si->size     = seg->len;
   si->foffset  = seg->offset;
   si->filename = VG_(arena_strdup)(VG_AR_SYMTAB, seg->filename);

   si->ref = 1;

   si->seg = seg;

   si->next = segInfo;
   segInfo = si;

   if (symtabd_send != -1) {
      Char cmd = VG_SYMTABD_READ_SEG_SYMBOLS;
      size_t len = strlen(seg->filename);
      UInt handle;

      checkWrite(symtabd_send, &cmd, 1, "command");
      checkWrite(symtabd_send, seg, sizeof *seg, "segment");
      checkWrite(symtabd_send, &len, sizeof len, "filename length");
      checkWrite(symtabd_send, seg->filename, len, "filename");
      checkRead(symtabd_recv, &handle, sizeof handle, "symtab handle");
      si->handle = handle;
   }

#if 0
   /* do redirects */
   resolve_seg_redirs( si );
#endif

   VGP_POPCC(VgpReadSyms);

   return si;
}

void
VG_(unload_symbols)( Addr start, UInt length )
{
   SegInfo *prev, *curr;

   prev = NULL;
   curr = segInfo;
   while (True) {
      if (curr == NULL) break;
      if (start == curr->start) break;
      prev = curr;
      curr = curr->next;
   }
   if (curr == NULL) {
      VGP_POPCC(VgpReadSyms);
      return;
   }

   vg_assert(prev == NULL || prev->next == curr);

   if (prev == NULL) {
      segInfo = curr->next;
   } else {
      prev->next = curr->next;
   }

   freeSegInfo(curr);
   return;
}

void
VG_(symtab_incref)( SegInfo *si )
{
   vg_assert(si->ref > 0);
   si->ref++;
}

void
VG_(symtab_decref)( SegInfo *si, Addr a, UInt len )
{
   vg_assert(si->ref >= 1);
   if (--si->ref == 0)
      VG_(unload_symbols)(si->start, si->size);
}

const SegInfo* VG_(next_seginfo)(const SegInfo* seg)
{
   if (seg == NULL)
      return segInfo;
   return seg->next;
}

Addr VG_(seg_start)(const SegInfo* seg)
{
   return seg->start;
}

UInt VG_(seg_size)(const SegInfo* seg)
{
   return seg->size;
}

const UChar* VG_(seg_filename)(const SegInfo* seg)
{
   return seg->filename;
}

/* The whole point of this whole big deal: map a code address to a
   plausible symbol name.  Returns False if no idea; otherwise True.
   Caller supplies buf and nbuf.  If demangle is False, don't do
   demangling, regardless of vg_clo_demangle -- probably because the
   call has come from vg_what_fn_or_object_is_this. */

static Bool
get_fnname(Char cmd, Addr a, Char* buf, Int nbuf)
{
   Int len;
   char* s;

   if (symtabd_send == -1)
      return False;

   checkWrite(symtabd_send, &cmd, 1, "command");
   checkWrite(symtabd_send, &a, sizeof a, "address");
   checkRead(symtabd_recv, &len, sizeof len, "fnname length");
   if (len > 0) {
      s = VG_(malloc)(len + 1);
      checkRead(symtabd_recv, s, len, "fnname");
      s[len] = 0;
      VG_(strncpy_safely)( buf, s, nbuf );
      VG_(free)(s);

      return True;
   } else {
      return False;
   }
}

/* This is available to skins... always demangle C++ names,
   match anywhere in function, but don't show offsets. */
Bool VG_(get_fnname) ( Addr a, Char* buf, Int nbuf )
{
   return get_fnname(VG_SYMTABD_GET_FNNAME, a, buf, nbuf);
}

/* This is available to skins... always demangle C++ names,
   match anywhere in function, and show offset if nonzero. */
Bool VG_(get_fnname_w_offset) ( Addr a, Char* buf, Int nbuf )
{
   return get_fnname(VG_SYMTABD_GET_FNNAME_W_OFFSET, a, buf, nbuf);
}

/* This is available to skins... always demangle C++ names,
   only succeed if 'a' matches first instruction of function,
   and don't show offsets. */
Bool VG_(get_fnname_if_entry) ( Addr a, Char* buf, Int nbuf )
{
   return get_fnname(VG_SYMTABD_GET_FNNAME_IF_ENTRY, a, buf, nbuf);
}

/* This is only available to core... don't demangle C++ names,
   match anywhere in function, and don't show offsets. */
Bool VG_(get_fnname_nodemangle) ( Addr a, Char* buf, Int nbuf )
{
   return get_fnname(VG_SYMTABD_GET_FNNAME_NODEMANGLE, a, buf, nbuf);
}

/* Map a code address to the name of a shared object file or the executable.
   Returns False if no idea; otherwise True.  Doesn't require debug info.
   Caller supplies buf and nbuf. */
Bool VG_(get_objname) ( Addr a, Char* buf, Int nbuf )
{
   SegInfo* si;

   for (si = segInfo; si != NULL; si = si->next) {
      if (si->start <= a && a < si->start+si->size) {
         VG_(strncpy_safely)(buf, si->filename, nbuf);
         return True;
      }
   }
   return False;
}

/* Map a code address to a filename.  Returns True if successful.  */
Bool VG_(get_filename)( Addr a, Char* filename, Int n_filename )
{
   return get_fnname(VG_SYMTABD_GET_FILENAME, a, filename, n_filename);
}

/* Map a code address to a line number.  Returns True if successful. */
Bool VG_(get_linenum)( Addr a, UInt* lineno )
{
   Char cmd = VG_SYMTABD_GET_LINENUM;

   if (symtabd_send == -1)
      return False;

   checkWrite(symtabd_send, &cmd, 1, "command");
   checkWrite(symtabd_send, &a, sizeof a, "address");
   checkRead(symtabd_recv, lineno, sizeof *lineno, "linenum");
   if (*lineno != ~0)
      return True;
   else
      return False;
}

/* Map a code address to a (filename, line number) pair.  
   Returns True if successful.
*/
Bool VG_(get_filename_linenum)( Addr a, 
                                Char* filename, Int n_filename, 
                                UInt* lineno )
{
   Char cmd = VG_SYMTABD_GET_FILENAME_LINENUM;
   Int len;
   char* s;

   if (symtabd_send == -1)
      return False;

   checkWrite(symtabd_send, &cmd, 1, "command");
   checkWrite(symtabd_send, &a, sizeof a, "address");
   checkRead(symtabd_recv, &len, sizeof len, "filename length");
   if (len > 0) {
      s = VG_(malloc)(len + 1);
      checkRead(symtabd_recv, s, len, "filename");
      checkRead(symtabd_recv, lineno, sizeof *lineno, "linenum");
      s[len] = 0;
      VG_(strncpy_safely)( filename, s, n_filename );
      VG_(free)(s);
      return True;
   } else {
      return False;
   }
}


/* Set up some default redirects */
void
VG_(setup_code_redirect_table)( void )
{
}


/* Redirection machinery */
void
VG_(add_redirect_sym)(const Char *from_lib, const Char *from_sym,
		      const Char *to_lib, const Char *to_sym)
{
}

void
VG_(add_redirect_addr)(const Char *from_lib, const Char *from_sym,
		       Addr to_addr)
{
}

Addr
VG_(code_redirect)(Addr orig)
{
   return orig;
}

/*--------------------------------------------------------------------*/
/*--- end                                             vg_symtab2.c ---*/
/*--------------------------------------------------------------------*/
