/*--------------------------------------------------------------------*/
/*--- Management of symbols and debugging information.             ---*/
/*---                                                 vg_symtab2.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of Valgrind, an extensible x86 protected-mode
   emulator for monitoring program execution on x86-Unixes.

   Copyright (C) 2000-2004 Julian Seward 
      jseward@acm.org

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#include "vg_include.h"
#include "vg_symtabd.h"
#include "vg_symtab2.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>

Int    VG_(clo_verbosity)      = 1;
Bool   VG_(clo_profile)        = False;
Bool   VG_(clo_show_below_main) = False;
Bool   VG_(clo_trace_symtab)   = False;
Int    VG_(clo_logfile_fd)     = 1;
Bool   VG_(clo_demangle)       = True;

VgNeeds VG_(needs) = {
   .core_errors          = False,
   .skin_errors          = False,
   .libc_freeres         = False,
   .basic_block_discards = False,
   .shadow_regs          = False,
   .command_line_options = False,
   .client_requests      = False,
   .extended_UCode       = False,
   .syscall_wrapper      = False,
   .sanity_checks        = False,
   .data_syms	         = False,
   .shadow_memory        = False,
};

__attribute__ ((noreturn))
static void report_and_quit ( const Char* report )
{
   exit(1);
}

void VG_(exit)( Int status )
{
   exit(status);
}

__attribute__ ((noreturn))
static void assert_fail ( const Char* expr, const Char* name, const Char* report,
                          const Char* file, Int line, const Char* fn )
{
   static Bool entered = False;
   if (entered) 
      VG_(exit)(2);
   entered = True;
   VG_(printf)("\n%s: %s:%d (%s): Assertion `%s' failed.\n",
               name, file, line, fn, expr );
   report_and_quit(report);
}

void VG_(skin_assert_fail) ( const Char* expr, const Char* file, Int line, const Char* fn )
{
   assert_fail(expr, "vg_symtabd", "dfr@freebsd.org", 
               file, line, fn);
}

void VG_(core_assert_fail) ( const Char* expr, const Char* file, Int line, const Char* fn )
{
   assert_fail(expr, "valgrind", VG_BUGS_TO, file, line, fn);
}

__attribute__ ((noreturn))
static void panic ( Char* name, Char* report, Char* str )
{
   VG_(printf)("\n%s: the `impossible' happened:\n   %s\n", name, str);
   report_and_quit(report);
}

void VG_(core_panic) ( Char* str )
{
   panic("vg_symtabd", VG_BUGS_TO, str);
}

UInt VG_(printf) ( const char *format, ... )
{
   UInt ret;
   va_list vargs;

   va_start(vargs,format);
   ret = vprintf(format, vargs);
   va_end(vargs);

   return ret;
}

UInt
VG_(sprintf)( Char* buf, Char *format, ... )
{
   UInt ret;
   va_list vargs;

   va_start(vargs,format);
   ret = vsprintf(buf, format, vargs);
   va_end(vargs);

   return ret;
}

int
VG_(add_to_msg) ( Char *format, ... )
{
   int count;
   va_list vargs;
   va_start(vargs,format);
   count = vprintf(format, vargs );
   va_end(vargs);
   return count;
}

int VG_(vmessage) ( VgMsgKind kind, Char* format, va_list vargs )
{
   int count;
   count = VG_(start_msg) ( kind );
   count += vprintf(format, vargs );
   count += VG_(end_msg)();
   return count;
}

/* Send a simple single-part message. */
int VG_(message) ( VgMsgKind kind, Char* format, ... )
{
   int count;
   va_list vargs;
   va_start(vargs,format);
   count = VG_(vmessage) ( kind, format, vargs );
   va_end(vargs);
   return count;
}

int VG_(start_msg) ( VgMsgKind kind )
{
   Char c;
   switch (kind) {
      case Vg_UserMsg:       c = '='; break;
      case Vg_DebugMsg:      c = '-'; break;
      case Vg_DebugExtraMsg: c = '+'; break;
      case Vg_ClientMsg:     c = '*'; break;
      default:               c = '?'; break;
   }
   return VG_(add_to_msg)( "%c%c%d%c%c ", 
                           c,c, getpid(), c,c );
}


int VG_(end_msg) ( void )
{
   int count = 0;
   if (VG_(clo_logfile_fd) >= 0) {
      printf("\n");
      count = 1;
   }
   return count;
}

Bool VG_(is_empty_arena) ( ArenaId aid )
{
   return True;			/* XXX */
}

void*
VG_(arena_malloc)( ArenaId arena, Int nbytes )
{
   return malloc(nbytes);
}

void
VG_(arena_free)( ArenaId arena, void* ptr )
{
   free(ptr);
}

void*
VG_(arena_realloc)( ArenaId arena, void* ptr, Int alignment, Int size )
{
   return realloc(ptr, size);
}

Char* VG_(arena_strdup)( ArenaId aid, const Char* s)
{
   return strdup(s);
}

Bool VG_(isspace) ( Char c )
{
   return (c == ' ' || c == '\n' || c == '\t' || c == 0);
}

Bool VG_(isdigit) ( Char c )
{
   return (c >= '0' && c <= '9');
}

Int
VG_(strlen)( const Char* str )
{
   return strlen(str);
}

Int
VG_(strcmp)( const Char* s1, const Char* s2 )
{
   return strcmp(s1, s2);
}

Int
VG_(strncmp)( const Char* s1, const Char* s2, Int nmax )
{
   return strncmp(s1, s2, nmax);
}

Char*
VG_(strchr)( const Char* s, Char c )
{
   return strchr(s, c);
}

Char* VG_(strpbrk)( const Char* s, const Char* accept )
{
   return strpbrk(s, accept);
}

Char*
VG_(strcpy)( Char* dest, const Char* src )
{
   return strcpy(dest, src);
}

Char* VG_(strncpy)( Char* dest, const Char* src, Int ndest )
{
   return strncpy(dest, src, ndest);
}

Char*
VG_(strdup)( const Char* s)
{
   return strdup(s);
}

/* Copy bytes, not overrunning the end of dest and always ensuring
   zero termination. */
void VG_(strncpy_safely) ( Char* dest, const Char* src, Int ndest )
{
   Int i;
   vg_assert(ndest > 0);
   i = 0;
   dest[i] = 0;
   while (True) {
      if (src[i] == 0) return;
      if (i >= ndest-1) return;
      dest[i] = src[i];
      i++;
      dest[i] = 0;
   }
}

Char* VG_(strstr) ( const Char* haystack, Char* needle )
{
   Int n; 
   if (haystack == NULL)
      return NULL;
   n = VG_(strlen)(needle);
   while (True) {
      if (haystack[0] == 0) 
         return NULL;
      if (VG_(strncmp)(haystack, needle, n) == 0) 
         return (Char*)haystack;
      haystack++;
   }
}


void*
VG_(memcpy)( void *d, const void *s, Int sz )
{
   return memcpy(d, s, sz);
}

Int
VG_(memcmp)( const void* s1, const void* s2, Int n )
{
   return memcmp(s1, s2, n);
}

void*
VG_(memset)( void *s, Int c, Int sz )
{
   return memset(s, c, sz);
}

void VG_(invalidate_translations)( Addr start, UInt range, Bool unchain_blocks )
{
}

Addr VG_(search_transtab) ( Addr original_addr )
{
   return (Addr)0;
}


Long VG_(atoll) ( Char* str )
{
   Bool neg = False;
   Long n = 0;
   if (*str == '-') { str++; neg = True; };
   while (*str >= '0' && *str <= '9') {
      n = 10*n + (Long)(*str - '0');
      str++;
   }
   if (neg) n = -n;
   return n;
}


Long VG_(atoll36) ( UInt base, Char* str )
{
   Bool neg = False;
   Long n = 0;
   vg_assert(base >= 2 && base <= 36);
   if (*str == '-') { str++; neg = True; };
   while (True) {
      if (*str >= '0' 
          && *str <= (Char)('9' - (10 - base))) {
         n = base*n + (Long)(*str - '0');
      }
      else 
      if (base > 10 && *str >= 'A' 
          && *str <= (Char)('Z' - (36 - base))) {
         n = base*n + (Long)((*str - 'A') + 10);
      }
      else 
      if (base > 10 && *str >= 'a' 
          && *str <= (Char)('z' - (36 - base))) {
         n = base*n + (Long)((*str - 'a') + 10);
      }
      else {
	break;
      }
      str++;
   }
   if (neg) n = -n;
   return n;
}

/* Returns -1 on failure. */
Int VG_(open) ( const Char* pathname, Int flags, Int mode )
{  
   return open(pathname, flags, mode);
}

void VG_(close) ( Int fd )
{
   close(fd);
}

void* VG_(mmap)( void* start, UInt length, 
                 UInt prot, UInt flags, UInt fd, UInt offset)
{
   return mmap(start, length, prot, flags, fd, offset);
}

Int VG_(munmap)( void* start, Int length )
{
   return munmap(start, length);
}

Int VG_(stat) ( Char* file_name, struct vki_stat* buf )
{
   return stat(file_name, (struct stat*) buf);
}

// Generic shell sort.  Like stdlib.h's qsort().
void VG_(ssort)( void* base, UInt nmemb, UInt size,
                 Int (*compar)(void*, void*) )
{
   Int   incs[14] = { 1, 4, 13, 40, 121, 364, 1093, 3280,
                      9841, 29524, 88573, 265720,
                      797161, 2391484 };
   Int   lo = 0;
   Int   hi = nmemb-1;
   Int   i, j, h, bigN, hp;

   bigN = hi - lo + 1; if (bigN < 2) return;
   hp = 0; while (hp < 14 && incs[hp] < bigN) hp++; hp--;
   vg_assert(0 <= hp && hp < 14);

   #define SORT \
   for ( ; hp >= 0; hp--) { \
      h = incs[hp]; \
      for (i = lo + h; i <= hi; i++) { \
         ASSIGN(v,0, a,i); \
         j = i; \
         while (COMPAR(a,(j-h), v,0) > 0) { \
            ASSIGN(a,j, a,(j-h)); \
            j = j - h; \
            if (j <= (lo + h - 1)) break; \
         } \
         ASSIGN(a,j, v,0); \
      } \
   }

   // Specialised cases
   if (sizeof(UInt) == size) {

      #define ASSIGN(dst, dsti, src, srci) \
      (dst)[(dsti)] = (src)[(srci)];      
      
      #define COMPAR(dst, dsti, src, srci) \
      compar( (void*)(& (dst)[(dsti)]), (void*)(& (src)[(srci)]) )

      UInt* a = (UInt*)base;
      UInt  v[1];

      SORT;

   } else if (sizeof(UShort) == size) {
      UShort* a = (UShort*)base;
      UShort  v[1];

      SORT;

   } else if (sizeof(UChar) == size) {
      UChar* a = (UChar*)base;
      UChar  v[1];

      SORT;

      #undef ASSIGN
      #undef COMPAR

   // General case
   } else {
      char* a = base;
      char  v[size];      // will be at least 'size' bytes

      #define ASSIGN(dst, dsti, src, srci) \
      VG_(memcpy)( &dst[size*(dsti)], &src[size*(srci)], size );

      #define COMPAR(dst, dsti, src, srci) \
      compar( &dst[size*(dsti)], &src[size*(srci)] )

      SORT;

      #undef ASSIGN
      #undef COMPAR
   }
   #undef SORT
}

/* ---------------------------------------------------------------------
   A simple string matching routine, purloined from Hugs98.
      `*'    matches any sequence of zero or more characters
      `?'    matches any single character exactly 
      `\c'   matches the character c only (ignoring special chars)
      c      matches the character c only
   ------------------------------------------------------------------ */

/* Keep track of recursion depth. */
static Int recDepth;

static Bool string_match_wrk ( const Char* pat, const Char* str )
{
   vg_assert(recDepth >= 0 && recDepth < 500);
   recDepth++;
   for (;;) {
      switch (*pat) {
         case '\0' : return (*str=='\0');
         case '*'  : do {
                        if (string_match_wrk(pat+1,str)) {
                           recDepth--;
                           return True;
                        }
                     } while (*str++);
                     recDepth--;
                     return False;
         case '?'  : if (*str++=='\0') {
                        recDepth--;
                        return False;
                     }
                     pat++;
                     break;
         case '\\' : if (*++pat == '\0') {
                        recDepth--;
                        return False; /* spurious trailing \ in pattern */
                     }
                     /* falls through to ... */
         default   : if (*pat++ != *str++) {
                        recDepth--;
                        return False;
                     }
                     break;
      }
   }
}

Bool VG_(string_match) ( const Char* pat, const Char* str )
{
   Bool b;
   recDepth = 0;
   b = string_match_wrk ( pat, str );
   /*
   VG_(printf)("%s   %s   %s\n",
	       b?"TRUE ":"FALSE", pat, str);
   */
   return b;
}

static Int addrcmp(const void *ap, const void *bp)
{
   Addr a = *(Addr *)ap;
   Addr b = *(Addr *)bp;
   Int ret;

   if (a == b)
      ret = 0;
   else
      ret = (a < b) ? -1 : 1;

   return ret;
}

static Char *straddr(void *p)
{
   static Char buf[16];

   VG_(sprintf)(buf, "%p", *(Addr *)p);

   return buf;
}

static SkipList sk_segments = SKIPLIST_INIT(Segment, addr, addrcmp, straddr, VG_AR_CORE);

Segment *VG_(find_segment)(Addr a)
{
   return VG_(SkipList_Find)(&sk_segments, &a);
}

Segment *VG_(next_segment)(Segment *s)
{
   return VG_(SkipNode_Next)(&sk_segments, s);
}

Bool VG_(seg_overlaps)(const Segment *s, Addr p, UInt len)
{
   Addr se = s->addr+s->len;
   Addr pe = p+len;

   vg_assert(pe >= p);

   return (p < se && pe > s->addr);
}

static void
checkRead(Int fd, void* buf, size_t len, const char* s)
{
   if (read(fd, buf, len) != len) {
      char msg[256];
      sprintf(msg, "Error reading '%s'", s);
      VG_(core_panic)(msg);
   }
}

static void
checkWrite(Int fd, const void* buf, size_t len, const char* s)
{
   if (write(fd, buf, len) != len) {
      char msg[256];
      sprintf(msg, "Error writing '%s'", msg);
      VG_(core_panic)(msg);
   }
}

int
main(int argc, char** argv)
{
   Char cmd;
   Int recv_fd;
   Int send_fd;

   if (argc != 3)
      exit(1);

   recv_fd = atoi(argv[1]);
   send_fd = atoi(argv[2]);
   //VG_(printf)("vg_symtabd %d %d\n", recv_fd, send_fd);

   for (;;) {
      int n;

      n = read(recv_fd, &cmd, 1);
      if (n == 0)
	 break;

      if (n != 1) {
	 VG_(core_panic)("Can't read command");
      }

      //VG_(printf)("cmd %d\n", cmd);
      switch (cmd) {
      case VG_SYMTABD_QUIT:
	 exit(0);

      case VG_SYMTABD_INIT:
      {
	 /*
	  * Message format:
	  *	Int		clo_verbosity
	  *	Bool		clo_profile
	  *	Bool		clo_show_below_main
	  *	Bool		clo_trace_symtab
	  *	Int		clo_logfile_fd
	  *	Bool		clo_demangle
	  *	VgNeeds		needs
	  */
	 checkRead(recv_fd, &VG_(clo_verbosity),	sizeof(Int),
	    "clo_verbosity");
	 checkRead(recv_fd, &VG_(clo_profile),		sizeof(Bool),
	    "clo_profile");
	 checkRead(recv_fd, &VG_(clo_show_below_main),	sizeof(Bool),
	    "clo_show_below_main");
	 checkRead(recv_fd, &VG_(clo_trace_symtab),	sizeof(Bool),
	    "clo_trace_symtab");
	 checkRead(recv_fd, &VG_(clo_logfile_fd),	sizeof(Int),
	    "clo_logfile_fd");
	 checkRead(recv_fd, &VG_(needs),		sizeof(VgNeeds),
	    "needs");
	 break;
      }

      case VG_SYMTABD_READ_SEG_SYMBOLS:
      {
	 UInt len;
	 Segment seg, *sp;
	 char* filename;
	 SegInfo* seginfo;
	 
	 /*
	  * Message format:
	  *	Segment		segment details
	  *	UInt		filename length
	  *	char[...]	filename
	  *
	  * Reply format:
	  *	SegInfo*	symtol table handle
	  */
	 checkRead(recv_fd, &seg, sizeof seg,
		   "segment details");
	 checkRead(recv_fd, &len, sizeof len,
		   "segment filename length");
	 seg.filename = filename = malloc(len + 1);
	 checkRead(recv_fd, filename, len,
		   "segment filename");
	 filename[len] = 0;
	 seginfo = VG_(read_seg_symbols)(&seg);
	 checkWrite(send_fd, &seginfo, sizeof seginfo,
		    "seginfo result");
	 free(filename);
	 sp = VG_(SkipNode_Alloc)(&sk_segments);
	 *sp = seg;
	 sp->symtab = seginfo;
	 VG_(SkipList_Insert)(&sk_segments, sp);
	 break;
      }

      case VG_SYMTABD_FREE_SYMBOLS:
      {
	 /*
	  * Message format:
	  *	SegInfo*	symbol table handle
	  */
	 SegInfo* si;
	 Segment* seg;
	 checkRead(recv_fd, &si, sizeof si, "seginfo");
	 seg = VG_(SkipList_Remove)(&sk_segments, &si->start);
	 VG_(symtab_decref)(si, si->start, si->size);
	 VG_(SkipNode_Free)(&sk_segments, seg);
	 break;
      }

      case VG_SYMTABD_GET_FNNAME:
      case VG_SYMTABD_GET_FNNAME_W_OFFSET:
      case VG_SYMTABD_GET_FNNAME_IF_ENTRY:
      case VG_SYMTABD_GET_FNNAME_NODEMANGLE:
      case VG_SYMTABD_GET_FILENAME:
      case VG_SYMTABD_GET_FILENAME_LINENUM:
      {
	 /*
	  * Message format:
	  *	UInt		address
	  *
	  * Reply format:
	  *	UInt		fnname length
	  *	char[...]	fnname
	  *	UInt		lineno (only for get_filename_linenum)
	  */
	 Addr a;
	 Char buf[256];
	 UInt len;
	 Bool res = False;
	 UInt lineno;

	 checkRead(recv_fd, &a, sizeof a, "address");
	 switch (cmd) {
	 case VG_SYMTABD_GET_FNNAME:
	    res = VG_(get_fnname)(a, buf, sizeof buf);
	    break;
	 case VG_SYMTABD_GET_FNNAME_W_OFFSET:
	    res = VG_(get_fnname_w_offset)(a, buf, sizeof buf);
	    break;
	 case VG_SYMTABD_GET_FNNAME_IF_ENTRY:
	    res = VG_(get_fnname_if_entry)(a, buf, sizeof buf);
	    break;
	 case VG_SYMTABD_GET_FNNAME_NODEMANGLE:
	    res = VG_(get_fnname_nodemangle)(a, buf, sizeof buf);
	    break;
	 case VG_SYMTABD_GET_FILENAME:
	    res = VG_(get_filename)(a, buf, sizeof buf);
	    break;
	 case VG_SYMTABD_GET_FILENAME_LINENUM:
	    res = VG_(get_filename_linenum)(a, buf, sizeof buf, &lineno);
	    break;
	 }
	 if (res) {
	    len = strlen(buf);
	    checkWrite(send_fd, &len, sizeof len, "fnname length");
	    checkWrite(send_fd, buf, len, "fnname");
	    if (cmd == VG_SYMTABD_GET_FILENAME_LINENUM)
	       checkWrite(send_fd, &lineno, sizeof lineno, "lineno");
	 } else {
	    len = 0;
	    checkWrite(send_fd, &len, sizeof len, "fnname length");
	 }
	 break;
      }
      case VG_SYMTABD_GET_LINENUM:
      {
	 /*
	  * Message format:
	  *	UInt		address
	  *
	  * Reply format:
	  *	UInt		lineno (~0 for none)
	  */
	 Addr a;
	 UInt lineno;

	 checkRead(recv_fd, &a, sizeof a, "address");
	 if (VG_(get_linenum)(a, &lineno))
	    lineno = ~0;
	 checkWrite(send_fd, &lineno, sizeof lineno, "lineno");
	 break;
      }

      default:
	 VG_(printf)("Read message %d\n", cmd);
	 VG_(core_panic)("Unrecognised message");
      }
   }

   return 0;
}

/*--------------------------------------------------------------------*/
/*--- end                                             vg_symtabd.c ---*/
/*--------------------------------------------------------------------*/
